# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

"""
This package contains various command line interfaces for common
pymatgen functionality such as file conversion, etc. Entry points to these
interfaces are defined in setup.py.
"""
